
from . import BackupHandler
from ..entity.definition import BackupDefinition
from ..exceptions import ReadWriteException
from ..result import CommandExecutionResult


class AbstractDocker(BackupHandler):
    def _execute_in_container(self,
                              docker_bin: str,
                              container: str,
                              command: str,
                              definition: BackupDefinition,
                              allocate_pts=False,
                              interactive=False,
                              with_crypto=True,
                              stdin=None,
                              mode='backup') -> CommandExecutionResult:

        """ Executes any command inside of the container """

        opts = ''

        if allocate_pts:
            opts += ' -t'

        if interactive:
            opts += ' -i'

        method = self._pipe_factory.create_backup_command

        if mode == 'restore':
            method = self._pipe_factory.create_restore_command

        return self._execute_command(
            method(
                docker_bin + ' exec ' + opts + ' ' + container + ' /bin/sh -c "' + command.replace('"', '\\"') + '"',
                definition,
                with_crypto=with_crypto
            ),
            stdin=stdin
        )

    def backup_directories(self, docker_bin: str, container: str, paths: list,
                           definition: BackupDefinition) -> CommandExecutionResult:
        """ Performs a backup of multiple directories using TAR with gzip/xz/bz2 compression """

        return self._execute_in_container(
            docker_bin, container,
            definition.get_pack_cmd(paths),
            definition
        )

    def _spawn_temporary_container(self,
                                   docker_bin: str,
                                   origin_container: str,
                                   temp_image_name: str,
                                   temp_container_cmd: str):
        """ Runs a temporary container that has mounted volumes from other container """

        temp_container_name = origin_container + '_backup_' + self.generate_id()

        run_command = docker_bin + ' run -d --rm --volumes-from ' + origin_container + \
                                   ' --name ' + temp_container_name + \
                                   ' ' + temp_image_name + \
                                   ' /bin/sh -c "' + temp_container_cmd + '"'

        result = self._execute_command(
            self._pipe_factory.create_backup_command(
                run_command,
                self._get_definition(),
                with_crypto=False
            )
        )

        out, code = result.process.communicate()
        container_id = out.decode('utf-8').strip()

        if not container_id or (result.process.returncode != 0 and result.process.returncode is not None):
            raise ReadWriteException('Cannot run temporary docker container, please verify image and command. Output: '
                                     + container_id)

        return container_id

    def _assert_all_paths_exists(self,
                                 docker_bin: str,
                                 container: str,
                                 paths: list,
                                 definition: BackupDefinition):
        """ Multiple assertion for directory/files presence """

        self.assert_container_running(docker_bin, container)

        for path in paths:
            self._assert_path_exists_in_container(docker_bin, container, path, definition)

    def _assert_path_exists_in_container(self,
                                         path: str,
                                         docker_bin: str,
                                         container: str,
                                         definition):
        """ Checks if a single directory or file exists inside of a container """

        response = self._execute_in_container(
            docker_bin,
            container,
            '[ -e ' + path + ' ] || echo does-not-exist',
            definition,
            with_crypto=False
        )

        code = response.process.wait()

        if int(code) > 0:
            raise ReadWriteException('Non-zero exit code from command, check if the container name is valid')

        if "does-not-exist" in response.stdout.read().decode('utf-8'):
            raise ReadWriteException(
                'Path "' + path + '" does not exist in container "' + definition.get_container() + '"'
            )

    def assert_container_running(self, docker_bin: str, container: str):
        """ Checks if a docker container is running """

        response = self._execute_command(
            docker_bin + ' ps | grep "' + container + '"'
        )

        response.process.wait()
        output = (str(response.stdout.read()) + str(response.stderr.read())).lower()

        if "got permission denied while trying to connect" in output:
            raise ReadWriteException(
                'You do not have access rights to the docker daemon, shoudn\'t you use sudo in docker_bin?'
            )

        if "cannot connect" in output or "connection refused" in output:
            raise ReadWriteException(
                'Docker daemon seems to be not running, or you do not have access rights to access it'
            )

        if container not in output:
            raise ReadWriteException('Container seems to be not running, check docker ps')

        if response.process.returncode > 0:
            raise ReadWriteException('Command failed with non-zero exit code, probably the container is not running, output: ' + output)

    @staticmethod
    def _assert_container_name_present(out: CommandExecutionResult, container_name: str, operation_name: str):
        stdout = out.stdout.read().decode('utf-8').replace('"', '').replace("'", '').strip()
        container_name = container_name.replace('"', '').strip()

        if not stdout == container_name:
            raise ReadWriteException('Cannot ' + operation_name + ' container "' + container_name + '". ' + stdout)

    def _kill_container(self, docker_bin: str, container_name: str):
        out = self._execute_command(docker_bin + ' kill ' + container_name)
        self._assert_container_name_present(out, container_name, 'kill')

    def _stop_container(self, docker_bin: str, container_name: str):
        out = self._execute_command(docker_bin + ' stop ' + container_name)
        self._assert_container_name_present(out, container_name, 'stop')

    def _start_container(self, docker_bin: str, container_name: str):
        out = self._execute_command(docker_bin + ' start ' + container_name)
        self._assert_container_name_present(out, container_name, 'start')





import requests
import json
from time import sleep
from ..entity.definition import BackupDefinition


class NotifierInterface:
    config = {}

    def __init__(self, config: dict):
        self._set_config(config)

    def _set_config(self, config: dict):
        pass

    def backup_was_uploaded(self, definition: BackupDefinition) -> None:
        pass

    def starting_backup_restore(self, definition: BackupDefinition) -> None:
        pass

    def starting_backup_creation(self, definition: BackupDefinition) -> None:
        pass

    def backup_was_restored(self, definition: BackupDefinition) -> None:
        pass

    def failed_to_upload_backup(self, definition: BackupDefinition, reason) -> None:
        pass

    def failed_to_restore_backup(self, definition: BackupDefinition, reason) -> None:
        pass

    def exception_occurred(self, exception: BaseException):
        pass


class Notifier:
    _notifiers = []

    def __init__(self, notifiers: dict):
        self._notifiers = notifiers.items()

    def __getattr__(self, item):
        return self._create_proxy(item)

    def _create_proxy(self, item):
        def _proxy(*args, **kwargs):
            for handler_name, handler in _proxy.notifiers:
                method = getattr(handler, _proxy.item)
                method(*args, **kwargs)

        _proxy.item = item
        _proxy.notifiers = self._notifiers

        return _proxy


class SlackNotifier(NotifierInterface):
    """ Slack/Mattermost notifications """

    _timeout = 300
    _url = ''
    _max_retry_num = 3

    def __init__(self, config: dict):
        super().__init__(config)
        self._url = config['url']
        self._max_retry_num = int(config.get('max_retry_num', 3))
        self._timeout = int(config.get('connection_timeout', 300))

    def backup_was_uploaded(self, definition: BackupDefinition) -> None:
        self._send(':white_check_mark: Backup was uploaded for ' + str(definition))

    def starting_backup_restore(self, definition: BackupDefinition) -> None:
        self._send(':information_source: Restoring backup of ' + str(definition))

    def starting_backup_creation(self, definition: BackupDefinition) -> None:
        self._send(':information_source: Creating backup of ' + str(definition))

    def backup_was_restored(self, definition: BackupDefinition) -> None:
        self._send(':white_check_mark: Backup was restored for ' + str(definition))

    def failed_to_upload_backup(self, definition: BackupDefinition, reason) -> None:
        self._send(':x: Failed to upload the backup for ' + str(definition) + ', ' + str(reason))

    def failed_to_restore_backup(self, definition: BackupDefinition, reason) -> None:
        self._send(':x: Failed to restore the backup for ' + str(definition) + ', ' + str(reason))

    def exception_occurred(self, exception: BaseException):
        self._send(':bangbang: ' + str(exception))

    def _send(self, msg: str, retry_num: int = 0):
        try:
            response = requests.post(
                self._url, data=json.dumps({'text': msg}),
                headers={'Content-Type': 'application/json'},
                timeout=self._timeout
            )
            if response.status_code != 200:
                raise ValueError(
                    'Request to slack returned an error %s, the response is:\n%s'
                    % (response.status_code, response.text)
                )
        except Exception as e:
            if self._max_retry_num and retry_num < self._max_retry_num:
                self._send(msg, retry_num + 1)
                sleep(1)
                return

            print('During sending a notification an unrecoverable error occurred:', e)


class NotifierFactory:
    _SUPPORTED = {
        'slack': SlackNotifier,
        'mattermost': SlackNotifier,
        'none': NotifierInterface,
        'nothing': NotifierInterface,
        '': NotifierInterface,
        'void': NotifierInterface
    }

    @staticmethod
    def create(notifier_type: str, config: dict):
        if notifier_type not in NotifierFactory._SUPPORTED:
            raise Exception('Unsupported notifier type "' + notifier_type + '"')

        return NotifierFactory._SUPPORTED[notifier_type](config)



from . import BackupHandler
from ..result import CommandExecutionResult
from ..exceptions import ReadWriteException
from ..entity.definition import CommandOutputDefinition


class CommandOutputBackup(BackupHandler):

    def _get_definition(self) -> CommandOutputDefinition:
        return self._definition

    def _validate(self):
        pass

    def _read_import_stream(self) -> CommandExecutionResult:
        return self._execute_command(
            self._pipe_factory.create_backup_command(
                self._get_definition().get_command(),
                self._get_definition()
            )
        )

    def _write(self, stream) -> CommandExecutionResult:
        definition = self._get_definition()

        if not definition.get_restore_command():
            raise ReadWriteException('Restore command not defined, cannot restore')

        return self._execute_command(
            self._pipe_factory.create_restore_command(definition.get_restore_command(), definition),
            stdin=stream
        )





class ServerAccess:

    _url = ""    # type: str
    _token = ""  # type: str

    def __init__(self, url: str, token: str):
        self._url = url
        self._token = token

    @staticmethod
    def from_config(config: dict):
        return ServerAccess(config['url'], config['token'])

    def get_url(self):
        return self._url

    def get_token(self):
        return self._token

    def build_url(self, endpoint: str, with_token: bool, password: str = ''):
        url = self._url.rstrip('/') + '/' + endpoint.lstrip('/')

        qs = '?'

        if with_token:
            qs += '&_token=' + self._token

        if password:
            qs += '&password=' + password

        return url + qs.replace('?&', '?')





from ..entity.definition import BackupDefinition
from ..exceptions import InvalidRequestException, UnexpectedResponseException
from logging import Logger
import requests
from simplejson.errors import JSONDecodeError


class FileRepositoryClient:
    """ Http Client to the File Repository API """

    _logger = None  # type: Logger

    def __init__(self, _logger: Logger):
        self._logger = _logger

    def send(self, read_stream, definition: BackupDefinition):
        url = definition.get_access().build_url(
            '/repository/collection/' + definition.get_collection_id() + '/backup', True)
        response = requests.post(url, data=read_stream)

        try:
            _json = response.json()
        except JSONDecodeError:
            _json = {}

        self._logger.debug('Request: ' + str(url))
        self._logger.debug('response(' + response.text + ')')

        if response.status_code >= 400:
            raise InvalidRequestException(response.text, _json, _json.get('error_code', 0))

        return {
            'version': _json['version']['version'],
            'file_id': _json['version']['id'],
            'file_name': _json['version']['file']['filename']
        }

    def fetch(self, version: str, definition: BackupDefinition):
        url = definition.get_access().build_url(
            '/repository/collection/' + definition.get_collection_id() + '/backup/' + version,
            with_token=True)

        response = requests.get(url, stream=True)

        self._logger.debug('Request: ' + str(url))
        self._logger.debug('response_code=' + str(response.status_code))

        if response.status_code >= 400:
            self._logger.debug('response_code=' + str(response.text))
            raise InvalidRequestException(response.text, response.json(), response.json().get('error_code', 0))

        return response.raw

    def list_backups(self, definition: BackupDefinition) -> dict:
        url = definition.get_access().build_url(
            '/repository/collection/' + definition.get_collection_id() + '/backup',
            with_token=True)

        response = requests.get(url)
        json = response.json()
        versions = dict()

        if response.status_code >= 400:
            raise InvalidRequestException(response.text, response.json(), response.json().get('error_code', 0))

        if 'versions' not in json:
            raise UnexpectedResponseException(response.text)

        for k, version in json['versions'].items():
            versions['v' + str(version['details']['version'])] = {
                'id': version['details']['id'],
                'created': version['details']['creation_date']['date']
            }

        return versions



#!/usr/bin/env python3

import sys
import os
import argparse
import inspect

path = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.insert(0, path)

from bahubapp.service.configurationfactory import ConfigurationFactory
from bahubapp.app import Bahub
from bahubapp.service.logger import LoggerFactory
from bahubapp.service.errorhandler import ErrorHandlerService
from bahubapp.service.notifier import Notifier


def main():
    #
    # Arguments parsing
    #
    parser = argparse.ArgumentParser()
    parser.add_argument('options', metavar='options', type=str, nargs='+',
                        help='[backup/restore/list/recover/snapshot] [backup or recovery plan name]')

    parser.add_argument('--debug', help='Prints debugging messages', default=False, action="store_true")
    parser.add_argument('--uncensored', help='Do not remove credentials from logs', default=False, action="store_true")

    parser.add_argument('--config',
                        help='Path to the configuration file',
                        default=os.path.expanduser('~/.bahub.yaml'))

    parser.add_argument('--logs-path',
                        help='Logs path',
                        default=os.path.expanduser('/tmp'))

    parser.add_argument('--logs-file',
                        help='Log to a single file, instead of creating files by date',
                        default='')

    parser.description = 'Bahub - backup automation client for File Repository API'

    parsed = parser.parse_args()

    if 0 < len(parsed.options) < 2:
        print(' You need to specify two options eg. "backup some-name"')
        print('')
        print('Example usage:')
        print('  backup my_db_1')
        print('  restore my_db_1 latest')
        print('  restore my_db_1 v2')
        print('  list my_db_1')
        print('  recover my_recovery_plan_name')
        print('  snapshot my_recovery_plan_name')
        print('')
        sys.exit(1)

    if not os.path.isfile(parsed.config):
        print(' Configuration file "' + str(parsed.config) + '" does not exist')
        sys.exit(1)

    error_handler = None
    notifier = None

    try:
        config_factory = ConfigurationFactory(parsed.config, parsed.debug)
        notifier = Notifier(config_factory.get_notifiers())
        error_handler = ErrorHandlerService(config_factory.get_error_handlers())

        app = Bahub(
            factory=config_factory,
            options={
                'options': parsed.options,
                'debug': parsed.debug,
                'config': parsed.config
            },
            uncensored=parsed.uncensored,
            logger=LoggerFactory.create(parsed.debug, parsed.logs_path, parsed.logs_file),
            notifier=notifier
        )

        app.run_controller(parsed.options[0], parsed.options[1], parsed.debug, parsed.options)

    except Exception as e:
        if parsed.debug:
            raise e

        if error_handler:
            error_handler.record_exception(e)

        if notifier:
            notifier.exception_occurred(e)

        print(e)
        sys.exit(1)


if __name__ == "__main__":
    main()



from ..exceptions import ConfigurationFactoryException


class Encryption:
    """ Cryptography support (using OpenSSL command implementation) """

    _supported_crypto = [
        'aes-256-cbc', 'aes-256-ecb', 'aes-128-cbc', 'aes-128-cbc', 'des-ecb', 'des-cbc'
    ]

    _passphrase = ""    # type: str
    _method = ""        # type: str

    _encrypt_cmd = "openssl enc -%method% -pass pass:%pass%"
    _decrypt_cmd = "openssl enc -d -%method% -pass pass:%pass%"

    def __init__(self, passphrase: str, method: str, _encrypt_cmd: str = '', _decrypt_cmd: str = ''):
        self._passphrase = passphrase
        self._method = method

        if _encrypt_cmd:
            self._encrypt_cmd = _encrypt_cmd

        if _decrypt_cmd:
            self._decrypt_cmd = _decrypt_cmd

        if method and method not in self._supported_crypto:
            raise ConfigurationFactoryException('Crypto "' + method + '" is not supported. Please use one of: ' +
                                                str(self._supported_crypto))

    @staticmethod
    def from_config(config: dict):
        return Encryption(config['passphrase'], config['method'],
                          config.get('encrypt_cmd', ''), config.get('decrypt_cmd', ''))

    def get_passphrase(self) -> str:
        return self._passphrase

    def get_method(self) -> str:
        return self._method

    def should_use_crypto(self):
        return self.get_method() is not ""

    def create_encrypt_command(self):
        return self._encrypt_cmd \
            .replace('%method%', self.get_method()) \
            .replace('%pass%', self.get_passphrase())

    def create_decrypt_command(self):
        return self._decrypt_cmd \
            .replace('%method%', self.get_method()) \
            .replace('%pass%', self.get_passphrase())





from ..entity.definition import BackupDefinition
from . import AbstractController


class BackupController(AbstractController):

    def perform(self, definition_name: str):
        return self._perform_backup(
            definition=self._definition_factory.get_definition(definition_name)
        )

    def _perform_backup(self, definition: BackupDefinition):
        handler = self._init_handler(definition)
        self._notifier.starting_backup_creation(definition)

        try:
            result = handler.perform_backup()
            handler.close()

            self._notifier.backup_was_uploaded(definition)

        except KeyboardInterrupt:
            handler.close()
            result = ""

        except Exception as e:
            handler.close()
            self._notifier.failed_to_upload_backup(definition, e)

            raise e

        return result




from . import BackupHandler
from ..result import CommandExecutionResult
from ..entity.definition import LocalFileDefinition
import os


class LocalFileBackup(BackupHandler):

    def _get_definition(self) -> LocalFileDefinition:
        return self._definition

    def _validate(self):
        for path in self._get_definition().get_paths():
            if not os.path.exists(path):
                raise Exception('Path "' + path + '" does not exist')

    def _read_import_stream(self) -> CommandExecutionResult:
        """ Read from local directory and return as a TAR-OPENSSL stream """

        tar_cmd = self._get_definition().get_pack_cmd(self._get_definition().get_paths())

        return self._execute_command(
            self._pipe_factory.create_backup_command(tar_cmd, self._get_definition())
        )

    def _write(self, stream) -> CommandExecutionResult:
        """ Write to a local directory - unpack a TAR archive """

        return self._execute_command(
            self._pipe_factory.create_restore_command(
                self._get_definition().get_unpack_cmd(),
                self._get_definition()
            ),
            stdin=stream
        )





from ..entity.definition import BackupDefinition
from . import AbstractController


class RestoreController(AbstractController):
    """ Selects a handler and calls for a restore of a remote backup """

    def perform(self, definition_name: str, version: str):
        return self._perform_restore(
            definition=self._definition_factory.get_definition(definition_name),
            version=version
        )

    def _perform_restore(self, definition: BackupDefinition, version: str):
        handler = self._init_handler(definition)
        self._logger.info('Restoring version ' + version)
        self._notifier.starting_backup_restore(definition)

        try:
            result = handler.perform_restore(version)
            handler.close()

            self._notifier.backup_was_restored(definition)

        except KeyboardInterrupt:
            handler.close()
            result = ""

        except Exception as e:
            handler.close()
            self._notifier.failed_to_restore_backup(definition, e)

            raise e

        return result





from ..handler.dockervolumebackup import DockerVolumeHotBackup, DockerVolumeBackup
from ..handler.localfilebackup import LocalFileBackup
from ..handler.commandoutputbackup import CommandOutputBackup
from ..handler.dockeroutputbackup import DockerCommandOutputBackup
from ..handler.mysqlbackup import MySQLBackup


class HandlersMapping:
    _mapping = {
        'docker_hot_volumes': DockerVolumeHotBackup,
        'docker_volumes': DockerVolumeBackup,
        'mysql': MySQLBackup,
        'docker_output': DockerCommandOutputBackup,
        'command_output': CommandOutputBackup,
        'directory': LocalFileBackup
    }

    def get(self, name: str):
        """ Resolves "type" configuration key into object, on error throws KeyError """

        return self._mapping[name]

    def has_handler(self, name: str) -> bool:
        return name in self._mapping



from ..entity.definition import BackupDefinition
from ..service.client import FileRepositoryClient
from ..service.pipefactory import PipeFactory
from ..exceptions import ReadWriteException
from ..result import CommandExecutionResult
from logging import Logger
import string
import random
import subprocess
from shutil import copyfileobj


class BackupHandler:
    """ Manages the process of backup and restore, interacts with different sources of backup data using adapters """

    _client = None  # type: FileRepositoryClient
    _pipe_factory = None  # type: PipeFactory
    _logger = None  # type: Logger
    _definition = None

    def __init__(self,
                 _client: FileRepositoryClient,
                 _pipe_factory: PipeFactory,
                 _logger: Logger,
                 _definition: BackupDefinition):
        self._client = _client
        self._pipe_factory = _pipe_factory
        self._logger = _logger
        self._definition = _definition

    def perform_backup(self):
        self._validate()
        self._prevalidate_if_the_command_not_dies_early()

        response = self._read_import_stream()

        if response.return_code != 0 and response.return_code is not None:
            raise ReadWriteException('Backup source read error, use --debug and retry to investigate')

        upload_response = self._client.send(response.stdout, self._get_definition())
        response.process.wait(15)
        response.stdout.close()

        return upload_response

    def perform_restore(self, version: str):
        response = self._write(
            self._read_from_storage(version)
        )

        response.process.wait()
        self._logger.info('Waiting for process to finish')

        if response.return_code is not None and response.return_code > 0:
            raise ReadWriteException('Cannot write files to disk while restoring from backup. Errors: '
                                     + str(response.stderr.read().decode('utf-8')))

        self._logger.info('No errors found, sending success information')

        return '{"status": "OK"}'

    def close(self):
        self._logger.info('Finishing the process')
        self._close()

    def _get_definition(self) -> BackupDefinition:
        return self._definition

    def _execute_command(self, command: str, stdin=None) -> CommandExecutionResult:
        """
        Executes a command on local machine, returning stdout as a stream, and streaming in the stdin (optionally)
        """

        self._logger.debug('shell(' + command + ')')

        process = subprocess.Popen(command,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   stdin=subprocess.PIPE if stdin else None,
                                   executable='/bin/bash',
                                   shell=True)

        if stdin:
            self._logger.info('Copying stdin to process')

            try:
                copyfileobj(stdin, process.stdin)
            except BrokenPipeError:
                process.kill()

                raise ReadWriteException(
                    'Cannot write to process, broken pipe occurred, probably a tar process died. '
                    + str(process.stderr.read().decode('utf-8')) + ' ' + str(process.stdout.read().decode('utf-8'))
                )

            process.stdin.close()

        return CommandExecutionResult(process.stdout, process.stderr, process.returncode, process)

    def _prevalidate_if_the_command_not_dies_early(self):
        """ Validate if the command really exports the data, does not end up with an error """

        response = self._read_import_stream()
        response.stdout.read(1024)

        response.process.kill()
        response.process.wait(15)

        if response.process.returncode > 0:
            raise ReadWriteException(
                'The process exited with incorrect code, try to verify the command in with --debug switch'
            )

    def _validate(self):
        raise Exception('_validate() not implemented for handler')

    def _read_import_stream(self) -> CommandExecutionResult:
        """ TAR output or file stream buffered from ANY source for example """
        raise Exception('_read_import_stream() not implemented for handler')

    def _write(self, stream) -> CommandExecutionResult:
        """ A file stream or tar output be written into the storage. May be OpenSSL encoded, depends on definition """
        raise Exception('_write() not implemented for handler')

    def _read_from_storage(self, version: str):
        return self._client.fetch(version, self._get_definition())

    def _close(self):
        pass

    @staticmethod
    def generate_id(size=6, chars=string.ascii_uppercase + string.digits):
        return ''.join(random.choice(chars) for _ in range(size))




from .access import ServerAccess
from .encryption import Encryption


class BackupDefinition:
    """ Configuration definition containing credentials, tokens, ids """

    _access = None       # type: ServerAccess
    _type = ""           # type: str
    _encryption = None   # type: Encryption
    _collection_id = ""  # type: str
    _tar_pack_cmd = 'tar -czf %stdin% %paths%'
    _tar_unpack_cmd = 'tar -xzf %stdin% %target%'
    _name = ""

    def __init__(self, access: ServerAccess, _type: str, collection_id: str, encryption: Encryption,
                 tar_pack_cmd: str, tar_unpack_cmd: str, name: str):
        self._access = access
        self._type = _type
        self._encryption = encryption
        self._collection_id = collection_id
        self._tar_pack_cmd = tar_pack_cmd
        self._tar_unpack_cmd = tar_unpack_cmd
        self._name = name

    @staticmethod
    def from_config(config: dict, name: str):
        return BackupDefinition(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', BackupDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', BackupDefinition._tar_unpack_cmd),
            name=name
        )

    def get_pack_cmd(self, paths: list):
        return self._tar_pack_cmd \
            .replace('%stdin%', '-') \
            .replace('%paths%', self.join_paths_quotes_into_string(paths))

    def get_unpack_cmd(self):
        return self._tar_unpack_cmd \
            .replace('%target%', '-C /') \
            .replace('%stdin%', '-')

    @staticmethod
    def join_paths_quotes_into_string(paths: list):
        joined = ""

        for path in paths:
            joined += '"' + path + '" '

        return joined

    def get_access(self) -> ServerAccess:
        return self._access

    def get_encryption(self) -> Encryption:
        return self._encryption

    def get_type(self) -> str:
        return self._type

    def get_collection_id(self) -> str:
        return self._collection_id

    def __repr__(self):
        return 'Definition<name=' + self._name + ',collection_id=' + str(self.get_collection_id()) + '>'


class DockerVolumesDefinition(BackupDefinition):
    _container = ""
    _docker_bin = ""
    _paths = ""

    def __init__(self,
                 access: ServerAccess,
                 _type: str,
                 collection_id: str,
                 encryption: Encryption,
                 tar_pack_cmd: str,
                 tar_unpack_cmd: str,
                 container: str,
                 docker_bin: str,
                 paths: list,
                 name: str):

        super().__init__(access, _type, collection_id, encryption, tar_pack_cmd, tar_unpack_cmd, name)
        self._container = container
        self._docker_bin = docker_bin
        self._paths = paths

    def get_container(self) -> str:
        return self._container

    def get_docker_bin(self) -> str:
        return self._docker_bin

    def get_paths(self) -> list:
        return self._paths

    @staticmethod
    def from_config(config: dict, name: str):
        return DockerVolumesDefinition(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', BackupDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', BackupDefinition._tar_unpack_cmd),
            container=config['container'],
            docker_bin=config.get('docker_bin', 'sudo docker'),
            paths=config['paths'],
            name=name
        )

    def __repr__(self):
        return 'Definition<name=' + self._name + ',collection_id=' + str(self.get_collection_id()) + ',docker_container=' + \
               str(self.get_container()) + '>'


class DockerOfflineVolumesDefinition(DockerVolumesDefinition):
    _container = ''
    _docker_bin = ''
    _paths = ''
    _temp_image_name = ''
    _temp_cmd = ''

    def __init__(self,
                 access: ServerAccess,
                 _type: str,
                 collection_id: str,
                 encryption: Encryption,
                 tar_pack_cmd: str,
                 tar_unpack_cmd: str,
                 container: str,
                 docker_bin: str,
                 paths: list,
                 temp_image_name: str,
                 temp_cmd: str,
                 name: str):
        super().__init__(access, _type, collection_id, encryption, tar_pack_cmd,
                         tar_unpack_cmd, container, docker_bin, paths, name)

        self._container = container
        self._docker_bin = docker_bin
        self._paths = paths
        self._temp_image_name = temp_image_name
        self._temp_cmd = temp_cmd

    def get_temp_image_name(self):
        return self._temp_image_name

    def get_temp_cmd(self):
        return self._temp_cmd

    @staticmethod
    def from_config(config: dict, name: str):
        return DockerOfflineVolumesDefinition(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', BackupDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', BackupDefinition._tar_unpack_cmd),
            container=config['container'],
            docker_bin=config.get('docker_bin', 'sudo docker'),
            paths=config['paths'],
            temp_image_name=config.get('temp_image_name', 'alpine:3.9'),
            temp_cmd=config.get('temp_image_cmd', 'apk add --update xz bzip2 && sleep 3600'),
            name=name
        )


class DockerOutputDefinition(BackupDefinition):
    _container = ""
    _command = ""
    _restore_command = ""
    _docker_bin = ""

    def __init__(self,
                 access: ServerAccess,
                 _type: str,
                 collection_id: str,
                 encryption: Encryption,
                 tar_pack_cmd: str,
                 tar_unpack_cmd: str,
                 container: str,
                 command: str,
                 restore_command: str,
                 docker_bin: str,
                 name: str):

        super().__init__(access, _type, collection_id, encryption, tar_pack_cmd, tar_unpack_cmd, name)
        self._container = container
        self._command = command
        self._docker_bin = docker_bin if docker_bin else "sudo docker"
        self._restore_command = restore_command

    def get_command(self) -> str:
        return self._command

    def get_restore_command(self) -> str:
        return self._restore_command

    def get_container(self) -> str:
        return self._container

    def get_docker_bin(self) -> str:
        return self._docker_bin

    @staticmethod
    def from_config(config: dict, name: str):
        return DockerOutputDefinition(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', BackupDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', BackupDefinition._tar_unpack_cmd),
            container=config['container'],
            command=config['command'],
            restore_command=config.get('restore_command', ''),
            docker_bin=config.get('docker_bin'),
            name=name
        )

    def __repr__(self):
        return 'Definition<name=' + self._name + ',collection_id=' + \
               str(self.get_collection_id()) + ',docker_container=' + \
               str(self.get_container()) + '>'


class MySQLDefinition(BackupDefinition):
    _host = "localhost"
    _port = 3306
    _user = "root"
    _password = "root"
    _database = ""
    _docker_bin = "sudo docker"
    _container = ""
    _mysql_dump_cmd = 'mysqldump --skip-lock-tables -u %user% -P %port% -p%password% -h %host% %database%'
    _mysql_restore_cmd = 'mysql -u %user% -p%password% -h %host% -P %port% %database% %query_subcmd%'
    _mysql_query_subcmd = ' -e "%query%"'

    def __init__(self,
                 access: ServerAccess,
                 _type: str,
                 collection_id: str,
                 encryption: Encryption,
                 tar_pack_cmd: str,
                 tar_unpack_cmd: str,
                 host: str,
                 port: int,
                 user: str,
                 password: str,
                 database: str,
                 docker_bin: str,
                 container: str,
                 name: str):

        super().__init__(access, _type, collection_id, encryption, tar_pack_cmd, tar_unpack_cmd, name)
        self._host = host
        self._port = port
        self._user = user
        self._password = password
        self._database = database
        self._docker_bin = docker_bin
        self._container = container

    def get_container(self) -> str:
        return self._container

    def get_docker_bin(self) -> str:
        return self._docker_bin

    def should_use_docker(self) -> bool:
        return self.get_container() != '' and self.get_docker_bin() != ''

    def get_database(self) -> str:
        return self._database if self._database else '--all-databases'

    def is_copying_all_databases(self) -> bool:
        return self.get_database() is "--all-databases"

    def get_mysql_dump_args(self) -> str:
        return self._fill_template(self._mysql_dump_cmd)

    def get_mysql_command(self, query: str = '', use_database: bool = True):
        subcmd = self._mysql_query_subcmd.replace('%query%', query) if query else ''

        return self._fill_template(self._mysql_restore_cmd, use_database=use_database) \
            .replace('%query_subcmd%', subcmd)

    def _fill_template(self, template: str, use_database: bool = True):
        return template\
            .replace('%host%', self._host) \
            .replace('%user%', self._user) \
            .replace('%port%', str(self._port)) \
            .replace('%database%', self._database if use_database else '') \
            .replace('%password%', self._password)

    @staticmethod
    def from_config(config: dict, name: str):
        return MySQLDefinition(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', BackupDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', BackupDefinition._tar_unpack_cmd),
            host=config['host'],
            port=int(config['port']),
            user=config['user'],
            password=config['password'],
            database=config['database'],
            docker_bin=config.get('docker_bin', 'sudo docker'),
            container=config.get('container', ''),
            name=name
        )

    def __repr__(self):
        return 'Definition<name=' + self._name + ',collection_id=' + \
               str(self.get_collection_id()) + ',docker_container=' + \
               str(self.get_container()) + ',sql_host=' + str(self._host) + '>'


class CommandOutputDefinition(BackupDefinition):
    _command = ""
    _restore_command = ""

    def __init__(self,
                 access: ServerAccess,
                 _type: str,
                 collection_id: str,
                 encryption: Encryption,
                 tar_pack_cmd: str,
                 tar_unpack_cmd: str,
                 command: str,
                 restore_command: str,
                 name: str):

        super().__init__(access, _type, collection_id, encryption, tar_pack_cmd, tar_unpack_cmd, name)
        self._command = command
        self._restore_command = restore_command

    def get_command(self) -> str:
        return self._command

    def get_restore_command(self) -> str:
        return self._restore_command

    @staticmethod
    def from_config(config: dict, name: str):
        return CommandOutputDefinition(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', BackupDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', BackupDefinition._tar_unpack_cmd),
            command=config['command'],
            restore_command=config.get('restore_command', ''),
            name=name
        )


class LocalFileDefinition(BackupDefinition):
    _paths = ""

    def __init__(self,
                 access: ServerAccess,
                 _type: str,
                 collection_id: str,
                 encryption: Encryption,
                 tar_pack_cmd: str,
                 tar_unpack_cmd: str,
                 paths: list,
                 name: str):

        super().__init__(access, _type, collection_id, encryption, tar_pack_cmd, tar_unpack_cmd, name)
        self._paths = paths

    def get_paths(self) -> list:
        return self._paths

    @staticmethod
    def from_config(config: dict, name: str):
        return LocalFileDefinition(
            access=config['access'],
            _type=config['type'],
            collection_id=config['collection_id'],
            encryption=config['encryption'],
            tar_pack_cmd=config.get('tar_pack_cmd', BackupDefinition._tar_pack_cmd),
            tar_unpack_cmd=config.get('tar_unpack_cmd', BackupDefinition._tar_unpack_cmd),
            paths=config['paths'],
            name=name
        )




from ..entity.definition import BackupDefinition


class PipeFactory:

    @staticmethod
    def create_backup_command(command: str, definition: BackupDefinition, with_crypto=True):
        piped = 'set -o pipefail; ' + command
        enc = definition.get_encryption()

        if with_crypto and enc.should_use_crypto():
            # append "-d" to decrypt the encrypted file
            piped += "| " + enc.create_encrypt_command()

        piped += '; exit $?'

        return piped

    @staticmethod
    def create_restore_command(command: str, definition: BackupDefinition, with_crypto=True):
        piped = ''
        enc = definition.get_encryption()

        if with_crypto and enc.should_use_crypto():
            piped += ' ' + enc.create_decrypt_command() + ' | '

        piped += command
        return piped




from ..entity.definition import BackupDefinition, \
    MySQLDefinition, DockerVolumesDefinition, DockerOfflineVolumesDefinition, \
    DockerOutputDefinition, CommandOutputDefinition, LocalFileDefinition


class DefinitionsMapping:
    _mapping = {
        'docker_hot_volumes': DockerVolumesDefinition,
        'docker_volumes': DockerOfflineVolumesDefinition,
        'mysql': MySQLDefinition,
        'docker_output': DockerOutputDefinition,
        'command_output': CommandOutputDefinition,
        'directory': LocalFileDefinition
    }

    @staticmethod
    def get(name: str) -> BackupDefinition:
        """ Resolves "type" configuration key into object, on error throws KeyError """

        return DefinitionsMapping._mapping[name]




import subprocess
from typing import IO


class CommandExecutionResult:
    stdout = None       # type: IO
    stderr = None       # type: IO
    return_code = None  # type: int
    process = None      # type: subprocess.Popen

    def __init__(self, stdout: IO, stderr: IO, returncode: int, process: subprocess.Popen):
        self.stdout = stdout
        self.stderr = stderr
        self.return_code = returncode
        self.process = process

    def __del__(self):
        self.process.kill()




from ..exceptions import ConfigurationFactoryException


class RecoveryPlan:
    _plan = []
    _policy = ""

    def __init__(self, plan, policy: str, all_available_services: list):
        self._plan = plan
        self._policy = policy

        if plan == "all" or plan == "*":
            self._plan = all_available_services

        self._validate_against(all_available_services)

    @staticmethod
    def from_config(config: dict, all_available_services: list):

        if config['policy'] not in ['restore-whats-possible', 'stop-on-first-error']:
            raise KeyError('Unsupported policy selected "' + str(config['policy']) + '"')

        return RecoveryPlan(config['definitions'], config['policy'], all_available_services)

    def get_planned_definitions_to_recover_in_order(self):
        return self._plan

    def on_error_should_continue(self) -> bool:
        return self._policy == 'restore-whats-possible'

    def _validate_against(self, all_available_services: list):
        for name in self._plan:
            if name not in all_available_services:
                raise ConfigurationFactoryException(
                    'Recovery plan refers to undefined backup definition "' + name + '"' +
                    ', possible options: ' + str(all_available_services)
                )




from .abstractdocker import AbstractDocker
from ..result import CommandExecutionResult
from ..exceptions import ReadWriteException
from ..entity.definition import DockerOutputDefinition


class DockerCommandOutputBackup(AbstractDocker):

    def _get_definition(self) -> DockerOutputDefinition:
        return self._definition

    def _validate(self):
        self.assert_container_running(
            self._get_definition().get_docker_bin(),
            self._get_definition().get_container()
        )

    def _read_import_stream(self):
        definition = self._get_definition()

        return self._execute_in_container(
            definition.get_docker_bin(),
            definition.get_container(),
            definition.get_command(),
            definition,
            allocate_pts=True
        )

    def _write(self, stream) -> CommandExecutionResult:
        definition = self._get_definition()

        if not definition.get_restore_command():
            raise ReadWriteException('Restore command not defined, cannot restore')

        return self._execute_in_container(
            definition.get_docker_bin(),
            definition.get_container(),
            definition.get_restore_command(),
            definition,
            interactive=True,
            stdin=stream,
            mode='restore'
        )



from .abstractdocker import AbstractDocker
from ..entity.definition import DockerVolumesDefinition, DockerOfflineVolumesDefinition
from ..result import CommandExecutionResult
from ..exceptions import ReadWriteException


class DockerVolumeHotBackup(AbstractDocker):
    """
    Backups a RUNNING container. For some applications it may be safe to backup a running container, for some not.
    See DockerVolumesBackup and adjust backup method to the application you want to keep safe.
    Gets into the container and makes a backup of directories into a single tar file.
    """

    def _get_definition(self) -> DockerVolumesDefinition:
        return self._definition

    def _validate(self):
        definition = self._get_definition()

        self.assert_container_running(
            definition.get_docker_bin(),
            definition.get_container()
        )

        for path in definition.get_paths():
            self._assert_path_exists_in_container(
                path,
                definition.get_docker_bin(),
                definition.get_container(),
                definition
            )

    def _read_import_stream(self) -> CommandExecutionResult:
        definition = self._get_definition()

        return self.backup_directories(
            definition.get_docker_bin(),
            definition.get_container(),
            definition.get_paths(),
            definition
        )

    def _write(self, stream) -> CommandExecutionResult:
        """ Write to a local directory - unpack a TAR archive """

        definition = self._get_definition()

        return self._execute_in_container(
            definition.get_docker_bin(),
            definition.get_container(),
            definition.get_unpack_cmd(),
            definition,
            stdin=stream,
            mode='restore',
            interactive=True
        )


class DockerVolumeBackup(AbstractDocker):
    """
    Offline docker container backup. Runs a new container mounting volumes of other container and performs a backup
    of those mounted volumes. Fully secure option for all kind of applications, as the applications are shut down for
    a moment.
    """

    _container_id = ""

    def _get_definition(self) -> DockerOfflineVolumesDefinition:
        return self._definition

    def _validate(self):
        # check if container exists (may be stopped)?
        pass

    def _read_import_stream(self):
        definition = self._get_definition()
        self._stop_origin_and_start_temporary_containers()

        self._logger.info('Performing backup of origin container in offline mode')
        return self.backup_directories(
            definition.get_docker_bin(),
            self._container_id,
            definition.get_paths(),
            definition
        )

    def _write(self, stream) -> CommandExecutionResult:
        definition = self._get_definition()
        self._stop_origin_and_start_temporary_containers()

        self._logger.info('Restoring backup to the temporary container through mounted volumes of origin container')
        return self._execute_in_container(
            definition.get_docker_bin(),
            self._container_id,
            definition.get_unpack_cmd(),
            definition,
            stdin=stream,
            mode='restore',
            interactive=True
        )

    def _stop_origin_and_start_temporary_containers(self):
        """ Stop origin container and start a temporary container """

        definition = self._get_definition()

        self._logger.info('Stopping origin container')
        self._stop_container(definition.get_docker_bin(), definition.get_container())

        self._logger.info('Spawning temporary container with volumes from origin container')
        self._container_id = self._spawn_temporary_container(
            definition.get_docker_bin(),
            definition.get_container(),
            definition.get_temp_image_name(),
            definition.get_temp_cmd()
        )

    def _close(self):
        definition = self._get_definition()

        try:
            self._logger.info('Killing temporary container')
            self._kill_container(definition.get_docker_bin(), self._container_id)

        except Exception:
            self._logger.warning('Cannot kill temporary container "' + self._container_id + '"')

        self._logger.info('Starting origin container')
        self._start_container(definition.get_docker_bin(), definition.get_container())




from .abstractdocker import AbstractDocker
from ..result import CommandExecutionResult
from ..entity.definition import MySQLDefinition
from ..exceptions import ReadWriteException


class MySQLBackup(AbstractDocker):
    def _get_definition(self) -> MySQLDefinition:
        return self._definition

    def _validate(self):
        pass

    def _read_import_stream(self):
        """
        Use MySQL dump command to extract data from database
        """
        return self._execute_command_in_proper_context(
            command=self._get_definition().get_mysql_dump_args() + ' | gzip',
            mode='backup'
        )

    def _write(self, stream) -> CommandExecutionResult:
        """
        Use MySQL shell util to import the database again
        """

        if self._get_definition().is_copying_all_databases():
            raise ReadWriteException('Cannot restore all databases at once, sorry it\'s not supported yet')

        self._logger.info('Not sending any DROP TABLE, mysqldump should already have "drop if exists" in dump')

        return self._execute_command_in_proper_context(
            command='gunzip | ' + self._get_definition().get_mysql_command(),
            mode='restore',
            stdin=stream
        )

    def _execute_command_in_proper_context(self, command: str, mode: str,
                                           with_crypto: bool = True,
                                           stdin=None) -> CommandExecutionResult:
        """
        Execute command in docker or on host
        """

        factory_method = self._pipe_factory.create_restore_command if mode == 'restore' else \
            self._pipe_factory.create_backup_command

        definition = self._get_definition()

        if definition.should_use_docker():
            return self._execute_in_container(
                definition.get_docker_bin(),
                definition.get_container(),
                command,
                definition,
                mode=mode,
                interactive=True,
                stdin=stdin
            )

        return self._execute_command(
            factory_method(
                command=command,
                definition=self._get_definition(),
                with_crypto=with_crypto
            ),
            stdin=stdin
        )

